"use strict";
/* --------------------------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License. See License.txt in the project root for license information.
 * ------------------------------------------------------------------------------------------ */
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var _Visitor_currentSourceFile;
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const typescripts_1 = require("./typescripts");
const path = require("path");
const LSPBaseTypes = new Set(['URI', 'DocumentUri', 'integer', 'uinteger', 'decimal']);
var MapKeyType;
(function (MapKeyType) {
    function is(value) {
        return value.kind === 'reference' || (value.kind === 'base' && (value.name === 'string' || value.name === 'integer' || value.name === 'DocumentUri' || value.name === 'URI'));
    }
    MapKeyType.is = is;
})(MapKeyType || (MapKeyType = {}));
var TypeInfo;
(function (TypeInfo) {
    function isNonLSPType(info) {
        return info.kind === 'base' && (info.name === 'void' || info.name === 'undefined' || info.name === 'never' || info.name === 'unknown');
    }
    TypeInfo.isNonLSPType = isNonLSPType;
    function isVoid(info) {
        return info.kind === 'base' && info.name === 'void';
    }
    TypeInfo.isVoid = isVoid;
    const baseSet = new Set(['null', 'void', 'string', 'boolean', 'URI', 'DocumentUri', 'integer', 'uinteger', 'decimal']);
    function asJsonType(info) {
        switch (info.kind) {
            case 'base':
                if (baseSet.has(info.name)) {
                    return { kind: 'base', name: info.name };
                }
                if (info.name === 'object') {
                    return { kind: 'reference', name: 'LSPAny' };
                }
                break;
            case 'reference':
                return { kind: 'reference', name: info.name };
            case 'array':
                return { kind: 'array', element: asJsonType(info.elementType) };
            case 'map':
                return { kind: 'map', key: asJsonType(info.key), value: asJsonType(info.value) };
            case 'union':
                return { kind: 'or', items: info.items.map(info => asJsonType(info)) };
            case 'intersection':
                return { kind: 'and', items: info.items.map(info => asJsonType(info)) };
            case 'tuple':
                return { kind: 'tuple', items: info.items.map(info => asJsonType(info)) };
            case 'literal':
                const literal = { properties: [] };
                for (const entry of info.items) {
                    const property = { name: entry[0], type: asJsonType(entry[1].type) };
                    const value = entry[1];
                    if (value.optional === true) {
                        property.optional = true;
                    }
                    if (value.documentation !== undefined) {
                        property.documentation = value.documentation;
                    }
                    if (value.since !== undefined) {
                        property.since = value.since;
                    }
                    if (value.proposed === true) {
                        property.proposed = true;
                    }
                    literal.properties.push(property);
                }
                return { kind: 'literal', value: literal };
            case 'stringLiteral':
                return { kind: 'stringLiteral', value: info.value };
            case 'integerLiteral':
                return { kind: 'integerLiteral', value: info.value };
            case 'booleanLiteral':
                return { kind: 'booleanLiteral', value: info.value };
        }
        throw new Error(`Can't convert type info ${JSON.stringify(info, undefined, 0)}`);
    }
    TypeInfo.asJsonType = asJsonType;
})(TypeInfo || (TypeInfo = {}));
var MessageDirection;
(function (MessageDirection) {
    MessageDirection.clientToServer = 'clientToServer';
    MessageDirection.serverToClient = 'serverToClient';
    MessageDirection.both = 'both';
    function is(value) {
        return value === MessageDirection.clientToServer || value === MessageDirection.serverToClient || value === MessageDirection.both;
    }
    MessageDirection.is = is;
})(MessageDirection || (MessageDirection = {}));
class Visitor {
    constructor(program) {
        _Visitor_currentSourceFile.set(this, void 0);
        this.program = program;
        this.typeChecker = this.program.getTypeChecker();
        this.symbols = new typescripts_1.Symbols(this.typeChecker);
        this.requests = [];
        this.notifications = [];
        this.structures = [];
        this.enumerations = [];
        this.typeAliases = [];
        this.symbolQueue = new Map();
        this.processedStructures = new Map();
    }
    get currentSourceFile() {
        if (__classPrivateFieldGet(this, _Visitor_currentSourceFile, "f") === undefined) {
            throw new Error(`Current source file not known`);
        }
        return __classPrivateFieldGet(this, _Visitor_currentSourceFile, "f");
    }
    async visitProgram() {
        for (const sourceFile of this.getSourceFilesToIndex()) {
            this.visit(sourceFile);
        }
    }
    async endVisitProgram() {
        while (this.symbolQueue.size > 0) {
            const toProcess = new Map(this.symbolQueue);
            for (const entry of toProcess) {
                const element = this.processSymbol(entry[0], entry[1]);
                if (element === undefined) {
                    throw new Error(`Can't create structure for type ${entry[0]}`);
                }
                else if (Array.isArray(element.properties)) {
                    this.structures.push(element);
                }
                else if (Array.isArray(element.values)) {
                    this.enumerations.push(element);
                }
                else {
                    this.typeAliases.push(element);
                }
                this.symbolQueue.delete(entry[0]);
                this.processedStructures.set(entry[0], entry[1]);
            }
        }
    }
    getMetaModel() {
        return {
            metaData: { version: '3.17.0' },
            requests: this.requests,
            notifications: this.notifications,
            structures: this.structures,
            enumerations: this.enumerations,
            typeAliases: this.typeAliases
        };
    }
    visit(node) {
        switch (node.kind) {
            case ts.SyntaxKind.SourceFile:
                this.doVisit(this.visitSourceFile, this.endVisitSourceFile, node);
                break;
            case ts.SyntaxKind.ModuleDeclaration:
                this.doVisit(this.visitModuleDeclaration, this.endVisitModuleDeclaration, node);
                break;
            default:
                this.doVisit(this.visitGeneric, this.endVisitGeneric, node);
                break;
        }
    }
    doVisit(visit, endVisit, node) {
        if (visit.call(this, node)) {
            node.forEachChild(child => this.visit(child));
        }
        endVisit.call(this, node);
    }
    visitGeneric() {
        return true;
    }
    endVisitGeneric() {
    }
    visitSourceFile(node) {
        __classPrivateFieldSet(this, _Visitor_currentSourceFile, node, "f");
        // The file `protocol.$.ts` contains all definitions for things
        // that are not reference through the protocol module but part of
        // the LSP specification. So treat the definitions as such. They all need
        // to start with a $ to make this clear.
        if (path.basename(node.fileName) === 'protocol.$.ts') {
            for (const statement of node.statements) {
                if (ts.isTypeAliasDeclaration(statement) && statement.name.getText()[0] === '$') {
                    this.visitTypeReference(statement);
                }
                if (ts.isVariableStatement(statement)) {
                    for (const declaration of statement.declarationList.declarations) {
                        if (declaration.name.getText()[0] !== '$' || declaration.initializer === undefined) {
                            continue;
                        }
                        const symbol = this.typeChecker.getSymbolAtLocation(declaration.initializer);
                        if (symbol === undefined) {
                            continue;
                        }
                        this.queueSymbol(symbol.getName(), symbol);
                    }
                }
            }
        }
        return true;
    }
    endVisitSourceFile() {
        __classPrivateFieldSet(this, _Visitor_currentSourceFile, undefined, "f");
    }
    visitModuleDeclaration(node) {
        const identifier = node.name.getText();
        // We have a request or notification definition.
        if (identifier.endsWith('Request')) {
            const request = this.visitRequest(node);
            if (request === undefined) {
                throw new Error(`Creating meta data for request ${identifier} failed.`);
            }
            else {
                this.requests.push(request);
            }
        }
        else if (identifier.endsWith('Notification')) {
            const notification = this.visitNotification(node);
            if (notification === undefined) {
                throw new Error(`Creating meta data for notification ${identifier} failed.`);
            }
            else {
                this.notifications.push(notification);
            }
        }
        return true;
    }
    visitRequest(node) {
        const symbol = this.typeChecker.getSymbolAtLocation(node.name);
        if (symbol === undefined) {
            return;
        }
        const type = symbol.exports?.get('type');
        if (type === undefined) {
            return;
        }
        const methodName = this.getMethodName(symbol, type);
        if (methodName === undefined) {
            return;
        }
        const requestTypes = this.getRequestTypes(type);
        if (requestTypes === undefined) {
            return;
        }
        requestTypes.param && this.queueTypeInfo(requestTypes.param);
        this.queueTypeInfo(requestTypes.result);
        this.queueTypeInfo(requestTypes.partialResult);
        this.queueTypeInfo(requestTypes.errorData);
        this.queueTypeInfo(requestTypes.registrationOptions);
        const asJsonType = (info) => {
            if (TypeInfo.isNonLSPType(info)) {
                return undefined;
            }
            return TypeInfo.asJsonType(info);
        };
        const result = {
            method: methodName,
            result: TypeInfo.isVoid(requestTypes.result) ? TypeInfo.asJsonType({ kind: 'base', name: 'null' }) : TypeInfo.asJsonType(requestTypes.result),
            messageDirection: this.getMessageDirection(symbol)
        };
        result.params = requestTypes.param !== undefined ? asJsonType(requestTypes.param) : undefined;
        result.partialResult = asJsonType(requestTypes.partialResult);
        result.errorData = asJsonType(requestTypes.errorData);
        result.registrationMethod = this.getRegistrationMethodName(symbol);
        result.registrationOptions = asJsonType(requestTypes.registrationOptions);
        this.fillDocProperties(node, result);
        return result;
    }
    visitNotification(node) {
        const symbol = this.typeChecker.getSymbolAtLocation(node.name);
        if (symbol === undefined) {
            return;
        }
        const type = symbol.exports?.get('type');
        if (type === undefined) {
            return;
        }
        const methodName = this.getMethodName(symbol, type);
        if (methodName === undefined) {
            return;
        }
        const notificationTypes = this.getNotificationTypes(type);
        if (notificationTypes === undefined) {
            return undefined;
        }
        notificationTypes.param && this.queueTypeInfo(notificationTypes.param);
        this.queueTypeInfo(notificationTypes.registrationOptions);
        const asJsonType = (info) => {
            if (TypeInfo.isNonLSPType(info)) {
                return undefined;
            }
            return TypeInfo.asJsonType(info);
        };
        const result = {
            method: methodName,
            messageDirection: this.getMessageDirection(symbol)
        };
        result.params = notificationTypes.param !== undefined ? asJsonType(notificationTypes.param) : undefined;
        result.registrationMethod = this.getRegistrationMethodName(symbol);
        result.registrationOptions = asJsonType(notificationTypes.registrationOptions);
        this.fillDocProperties(node, result);
        return result;
    }
    visitTypeReference(node) {
        const type = node.type;
        if (!ts.isTypeReferenceNode(type)) {
            return;
        }
        const symbol = this.typeChecker.getSymbolAtLocation(type.typeName);
        if (symbol === undefined) {
            return;
        }
        this.queueSymbol(type.typeName.getText(), symbol);
    }
    queueTypeInfo(typeInfo) {
        if (typeInfo.kind === 'reference') {
            this.queueSymbol(typeInfo.name, typeInfo.symbol);
        }
        else if (typeInfo.kind === 'array') {
            this.queueTypeInfo(typeInfo.elementType);
        }
        else if (typeInfo.kind === 'union' || typeInfo.kind === 'intersection' || typeInfo.kind === 'tuple') {
            typeInfo.items.forEach(item => this.queueTypeInfo(item));
        }
        else if (typeInfo.kind === 'map') {
            this.queueTypeInfo(typeInfo.key);
            this.queueTypeInfo(typeInfo.value);
        }
        else if (typeInfo.kind === 'literal') {
            typeInfo.items.forEach(item => this.queueTypeInfo(item.type));
        }
    }
    queueSymbol(name, symbol) {
        if (name !== symbol.getName()) {
            throw new Error(`Different symbol names [${name}, ${symbol.getName()}]`);
        }
        const existing = this.symbolQueue.get(name) ?? this.processedStructures.get(name);
        if (existing === undefined) {
            const aliased = typescripts_1.Symbols.isAliasSymbol(symbol) ? this.typeChecker.getAliasedSymbol(symbol) : undefined;
            if (aliased !== undefined && aliased.getName() !== symbol.getName()) {
                throw new Error(`The symbol ${symbol.getName()} has a different name than the aliased symbol ${aliased.getName()}`);
            }
            this.symbolQueue.set(name, aliased ?? symbol);
        }
        else {
            const left = typescripts_1.Symbols.isAliasSymbol(symbol) ? this.typeChecker.getAliasedSymbol(symbol) : symbol;
            const right = typescripts_1.Symbols.isAliasSymbol(existing) ? this.typeChecker.getAliasedSymbol(existing) : existing;
            if (this.symbols.createKey(left) !== this.symbols.createKey(right)) {
                throw new Error(`The type ${name} has two different declarations`);
            }
        }
    }
    endVisitModuleDeclaration(_node) {
    }
    getSourceFilesToIndex() {
        const result = [];
        for (const sourceFile of this.program.getSourceFiles()) {
            if (this.program.isSourceFileFromExternalLibrary(sourceFile) || this.program.isSourceFileDefaultLibrary(sourceFile)) {
                continue;
            }
            result.push(sourceFile);
        }
        return result;
    }
    getMethodName(namespace, type) {
        const method = namespace.exports?.get('method');
        let text;
        if (method !== undefined) {
            const declaration = this.getFirstDeclaration(method);
            if (declaration === undefined) {
                return undefined;
            }
            if (!ts.isVariableDeclaration(declaration)) {
                return undefined;
            }
            const initializer = declaration.initializer;
            if (initializer === undefined || (!ts.isStringLiteral(initializer) && !ts.isNoSubstitutionTemplateLiteral(initializer))) {
                return undefined;
            }
            text = initializer.getText();
        }
        else {
            const declaration = this.getFirstDeclaration(type);
            if (declaration === undefined) {
                return undefined;
            }
            if (!ts.isVariableDeclaration(declaration)) {
                return undefined;
            }
            const initializer = declaration.initializer;
            if (initializer === undefined || !ts.isNewExpression(initializer)) {
                return undefined;
            }
            const args = initializer.arguments;
            if (args === undefined || args.length < 1) {
                return undefined;
            }
            text = args[0].getText();
        }
        return this.removeQuotes(text);
    }
    getRegistrationMethodName(namespace) {
        const registrationMethod = namespace.exports?.get('registrationMethod');
        if (registrationMethod === undefined) {
            return undefined;
        }
        const declaration = this.getFirstDeclaration(registrationMethod);
        if (declaration === undefined || !ts.isVariableDeclaration(declaration) || declaration.initializer === undefined || !ts.isPropertyAccessExpression(declaration.initializer)) {
            return undefined;
        }
        const initializerSymbol = this.typeChecker.getSymbolAtLocation(declaration.initializer.name);
        if (initializerSymbol === undefined || initializerSymbol.valueDeclaration === undefined) {
            return undefined;
        }
        const valueDeclaration = initializerSymbol.valueDeclaration;
        if (!ts.isVariableDeclaration(valueDeclaration)) {
            return undefined;
        }
        if (valueDeclaration.initializer === undefined || (!ts.isStringLiteral(valueDeclaration.initializer) && !ts.isNoSubstitutionTemplateLiteral(valueDeclaration.initializer))) {
            return undefined;
        }
        return this.removeQuotes(valueDeclaration.initializer.getText());
    }
    getMessageDirection(namespace) {
        const errorMessage = `No message direction specified for request ${namespace.getName()}`;
        const messageDirection = namespace.exports?.get('messageDirection');
        if (messageDirection === undefined) {
            throw new Error(errorMessage);
        }
        const declaration = this.getFirstDeclaration(messageDirection);
        if (declaration === undefined || !ts.isVariableDeclaration(declaration) || declaration.initializer === undefined || !ts.isPropertyAccessExpression(declaration.initializer)) {
            throw new Error(errorMessage);
        }
        const initializerSymbol = this.typeChecker.getSymbolAtLocation(declaration.initializer.name);
        if (initializerSymbol === undefined || initializerSymbol.valueDeclaration === undefined) {
            throw new Error(errorMessage);
        }
        const valueDeclaration = initializerSymbol.valueDeclaration;
        if (!ts.isEnumMember(valueDeclaration)) {
            throw new Error(errorMessage);
        }
        if (valueDeclaration.initializer === undefined || !ts.isStringLiteral(valueDeclaration.initializer)) {
            throw new Error(errorMessage);
        }
        const value = this.removeQuotes(valueDeclaration.initializer.getText());
        if (!MessageDirection.is(value)) {
            throw new Error(errorMessage);
        }
        return value;
    }
    getRequestTypes(symbol) {
        const declaration = this.getFirstDeclaration(symbol);
        if (declaration === undefined) {
            return undefined;
        }
        if (!ts.isVariableDeclaration(declaration)) {
            return;
        }
        const initializer = declaration.initializer;
        if (initializer === undefined || !ts.isNewExpression(initializer)) {
            return undefined;
        }
        if (initializer.typeArguments === undefined) {
            return undefined;
        }
        const typeInfos = [];
        for (const typeNode of initializer.typeArguments) {
            const info = this.getTypeInfo(typeNode);
            if (info === undefined) {
                return undefined;
            }
            typeInfos.push(info);
        }
        if (typeInfos.length !== initializer.typeArguments.length) {
            return undefined;
        }
        switch (initializer.typeArguments.length) {
            case 4:
                return { result: typeInfos[0], partialResult: typeInfos[1], errorData: typeInfos[2], registrationOptions: typeInfos[3] };
            case 5:
                return { param: typeInfos[0], result: typeInfos[1], partialResult: typeInfos[2], errorData: typeInfos[3], registrationOptions: typeInfos[4] };
        }
        return undefined;
    }
    getNotificationTypes(symbol) {
        const declaration = this.getFirstDeclaration(symbol);
        if (declaration === undefined) {
            return undefined;
        }
        if (!ts.isVariableDeclaration(declaration)) {
            return;
        }
        const initializer = declaration.initializer;
        if (initializer === undefined || !ts.isNewExpression(initializer)) {
            return undefined;
        }
        if (initializer.typeArguments === undefined) {
            return undefined;
        }
        const typeInfos = [];
        for (const typeNode of initializer.typeArguments) {
            const info = this.getTypeInfo(typeNode);
            if (info === undefined) {
                return undefined;
            }
            typeInfos.push(info);
        }
        if (typeInfos.length !== initializer.typeArguments.length) {
            return undefined;
        }
        switch (initializer.typeArguments.length) {
            case 1:
                return { registrationOptions: typeInfos[0] };
            case 2:
                return { param: typeInfos[0], registrationOptions: typeInfos[1] };
        }
        return undefined;
    }
    getTypeInfo(typeNode, isLSPAny = false) {
        if (ts.isIdentifier(typeNode)) {
            const typeName = typeNode.text;
            if (LSPBaseTypes.has(typeName)) {
                return { kind: 'base', name: typeName };
            }
            const symbol = this.typeChecker.getSymbolAtLocation(typeNode);
            if (symbol === undefined) {
                return undefined;
            }
            return { kind: 'reference', name: typeName, symbol };
        }
        else if (ts.isTypeReferenceNode(typeNode)) {
            const typeName = ts.isIdentifier(typeNode.typeName) ? typeNode.typeName.text : typeNode.typeName.right.text;
            if (LSPBaseTypes.has(typeName)) {
                return { kind: 'base', name: typeName };
            }
            const symbol = this.typeChecker.getSymbolAtLocation(typeNode.typeName);
            if (symbol === undefined) {
                return undefined;
            }
            return { kind: 'reference', name: typeName, symbol };
        }
        else if (ts.isArrayTypeNode(typeNode)) {
            const elementType = this.getTypeInfo(typeNode.elementType);
            if (elementType === undefined) {
                return undefined;
            }
            return { kind: 'array', elementType: elementType };
        }
        else if (ts.isUnionTypeNode(typeNode)) {
            const items = [];
            for (const item of typeNode.types) {
                const typeInfo = this.getTypeInfo(item);
                if (typeInfo === undefined) {
                    return undefined;
                }
                // We need to remove undefined from LSP Any since
                // it is not a valid type on the wire
                if (isLSPAny && typeInfo.kind === 'base' && typeInfo.name === 'undefined') {
                    continue;
                }
                items.push(typeInfo);
            }
            return { kind: 'union', items };
        }
        else if (ts.isIntersectionTypeNode(typeNode)) {
            const items = [];
            for (const item of typeNode.types) {
                const typeInfo = this.getTypeInfo(item);
                if (typeInfo === undefined) {
                    return undefined;
                }
                items.push(typeInfo);
            }
            return { kind: 'intersection', items };
        }
        else if (ts.isTypeLiteralNode(typeNode)) {
            const type = this.typeChecker.getTypeAtLocation(typeNode);
            const info = this.typeChecker.getIndexInfoOfType(type, ts.IndexKind.String);
            if (info !== undefined) {
                const declaration = info.declaration;
                if (declaration === undefined || declaration.parameters.length < 1) {
                    return undefined;
                }
                const keyTypeNode = declaration.parameters[0].type;
                if (keyTypeNode === undefined) {
                    return undefined;
                }
                const key = this.getTypeInfo(keyTypeNode);
                const value = this.getTypeInfo(declaration.type);
                if (key === undefined || value === undefined) {
                    return undefined;
                }
                if (!MapKeyType.is(key)) {
                    return undefined;
                }
                return { kind: 'map', key: key, value: value };
            }
            else {
                // We can't directly ask for the symbol since the literal has no name.
                const type = this.typeChecker.getTypeAtLocation(typeNode);
                const symbol = type.symbol;
                if (symbol === undefined) {
                    return undefined;
                }
                if (symbol.members === undefined) {
                    return { kind: 'literal', items: new Map() };
                }
                const items = new Map();
                symbol.members.forEach((member) => {
                    if (!typescripts_1.Symbols.isProperty(member)) {
                        return;
                    }
                    const declaration = this.getDeclaration(member, ts.SyntaxKind.PropertySignature);
                    if (declaration === undefined || !ts.isPropertySignature(declaration) || declaration.type === undefined) {
                        throw new Error(`Can't parse property ${member.getName()} of structure ${symbol.getName()}`);
                    }
                    const propertyType = this.getTypeInfo(declaration.type);
                    if (propertyType === undefined) {
                        throw new Error(`Can't parse property ${member.getName()} of structure ${symbol.getName()}`);
                    }
                    const literalInfo = { type: propertyType, optional: typescripts_1.Symbols.isOptional(member) };
                    this.fillDocProperties(declaration, literalInfo);
                    items.set(member.getName(), literalInfo);
                });
                return { kind: 'literal', items };
            }
        }
        else if (ts.isTupleTypeNode(typeNode)) {
            const items = [];
            for (const item of typeNode.elements) {
                const typeInfo = this.getTypeInfo(item);
                if (typeInfo === undefined) {
                    return undefined;
                }
                items.push(typeInfo);
            }
            return { kind: 'tuple', items };
        }
        else if (ts.isTypeQueryNode(typeNode) && ts.isQualifiedName(typeNode.exprName)) {
            // Currently we only us the typeof operator to get to the type of a enum
            // value expressed by an or type (e.g. kind: typeof DocumentDiagnosticReportKind.full)
            // So we assume a qualifed name and turn it into a string literal type
            const typeNodeSymbol = this.typeChecker.getSymbolAtLocation(typeNode.exprName);
            if (typeNodeSymbol === undefined) {
                throw new Error(`Can't resolve symbol for right hand side of enum declaration`);
            }
            const declaration = this.getDeclaration(typeNodeSymbol, ts.SyntaxKind.VariableDeclaration);
            if (declaration === undefined || !ts.isVariableDeclaration(declaration) || declaration.initializer === undefined) {
                throw new Error(`Can't resolve variable declaration for right hand side of enum declaration`);
            }
            if (ts.isNumericLiteral(declaration.initializer)) {
                return { kind: 'integerLiteral', value: Number.parseInt(declaration.initializer.getText()) };
            }
            else if (ts.isStringLiteral(declaration.initializer)) {
                return { kind: 'stringLiteral', value: this.removeQuotes(declaration.initializer.getText()) };
            }
            return { kind: 'stringLiteral', value: typeNode.exprName.right.getText() };
        }
        else if (ts.isParenthesizedTypeNode(typeNode)) {
            return this.getTypeInfo(typeNode.type);
        }
        else if (ts.isLiteralTypeNode(typeNode)) {
            return this.getBaseTypeInfo(typeNode.literal);
        }
        return this.getBaseTypeInfo(typeNode);
    }
    getBaseTypeInfo(node) {
        switch (node.kind) {
            case ts.SyntaxKind.NullKeyword:
                return { kind: 'base', name: 'null' };
            case ts.SyntaxKind.UnknownKeyword:
                return { kind: 'base', name: 'unknown' };
            case ts.SyntaxKind.NeverKeyword:
                return { kind: 'base', name: 'never' };
            case ts.SyntaxKind.VoidKeyword:
                return { kind: 'base', name: 'void' };
            case ts.SyntaxKind.UndefinedKeyword:
                return { kind: 'base', name: 'undefined' };
            case ts.SyntaxKind.AnyKeyword:
                return { kind: 'base', name: 'any' };
            case ts.SyntaxKind.StringKeyword:
                return { kind: 'base', name: 'string' };
            case ts.SyntaxKind.NumberKeyword:
                return { kind: 'base', name: 'integer' };
            case ts.SyntaxKind.BooleanKeyword:
                return { kind: 'base', name: 'boolean' };
            case ts.SyntaxKind.StringLiteral:
                return { kind: 'stringLiteral', value: this.removeQuotes(node.getText()) };
            case ts.SyntaxKind.NumericLiteral:
                return { kind: 'integerLiteral', value: Number.parseInt(node.getText()) };
            case ts.SyntaxKind.ObjectKeyword:
                return { kind: 'base', name: 'object' };
        }
        return undefined;
    }
    processSymbol(name, symbol) {
        // We can't define LSPAny in the protocol right now due to TS issues.
        // So we predefine it and emit it.
        if (name === 'LSPAny') {
            this.typeAliases.push(PreDefined.LSPArray);
            return PreDefined.LSPAny;
        }
        if (name === 'LSPArray') {
            // LSP Array is never reference via a indirect reference from
            // a request or notification.
            return undefined;
        }
        if (name === 'LSPObject') {
            return PreDefined.LSPObject;
        }
        if (typescripts_1.Symbols.isInterface(symbol)) {
            const result = { name: name, properties: [] };
            const declaration = this.getDeclaration(symbol, ts.SyntaxKind.InterfaceDeclaration);
            if (declaration !== undefined && ts.isInterfaceDeclaration(declaration) && declaration.heritageClauses !== undefined) {
                const mixins = [];
                const extend = [];
                for (const clause of declaration.heritageClauses) {
                    for (const type of clause.types) {
                        if (ts.isIdentifier(type.expression)) {
                            const typeInfo = this.getTypeInfo(type.expression);
                            if (typeInfo === undefined || typeInfo.kind !== 'reference') {
                                throw new Error(`Can't create type info for extends clause ${type.expression.getText()}`);
                            }
                            if (Visitor.Mixins.has(typeInfo.name)) {
                                mixins.push(TypeInfo.asJsonType(typeInfo));
                            }
                            else {
                                extend.push(TypeInfo.asJsonType(typeInfo));
                            }
                            this.queueTypeInfo(typeInfo);
                        }
                    }
                }
                if (extend.length > 0) {
                    result.extends = extend;
                }
                if (mixins.length > 0) {
                    result.mixins = mixins;
                }
            }
            if (declaration !== undefined) {
                this.fillDocProperties(declaration, result);
            }
            this.fillProperties(result, symbol);
            return result;
        }
        else if (typescripts_1.Symbols.isTypeAlias(symbol)) {
            const declaration = this.getDeclaration(symbol, ts.SyntaxKind.TypeAliasDeclaration);
            if (declaration === undefined || !ts.isTypeAliasDeclaration(declaration)) {
                throw new Error(`No declaration found for type alias ${symbol.getName()}`);
            }
            if (ts.isTypeLiteralNode(declaration.type)) {
                // We have a single type literal node. So treat it as a structure
                const result = { name: name, properties: [] };
                this.fillProperties(result, this.typeChecker.getTypeAtLocation(declaration.type).symbol);
                this.fillDocProperties(declaration, result);
                return result;
            }
            else if (ts.isIntersectionTypeNode(declaration.type)) {
                const split = this.splitIntersectionType(declaration.type);
                if (split.rest.length === 0) {
                    const result = { name: name, properties: [] };
                    const mixins = [];
                    const extend = [];
                    for (const reference of split.references) {
                        const typeInfo = this.getTypeInfo(reference);
                        if (typeInfo === undefined || typeInfo.kind !== 'reference') {
                            throw new Error(`Can't create type info for type reference ${reference.getText()}`);
                        }
                        if (Visitor.Mixins.has(typeInfo.name)) {
                            mixins.push(TypeInfo.asJsonType(typeInfo));
                        }
                        else {
                            extend.push(TypeInfo.asJsonType(typeInfo));
                        }
                        this.queueTypeInfo(typeInfo);
                    }
                    if (extend.length > 0) {
                        result.extends = extend;
                    }
                    if (mixins.length > 0) {
                        result.mixins = mixins;
                    }
                    if (split.literal !== undefined) {
                        this.fillProperties(result, this.typeChecker.getTypeAtLocation(split.literal).symbol);
                    }
                    this.fillDocProperties(declaration, result);
                    return result;
                }
            }
            const target = this.getTypeInfo(declaration.type, name === 'LSPAny');
            if (target === undefined) {
                throw new Error(`Can't resolve target type for type alias ${symbol.getName()}`);
            }
            const namespace = this.getDeclaration(symbol, ts.SyntaxKind.ModuleDeclaration);
            if (namespace !== undefined && symbol.declarations !== undefined && symbol.declarations.length === 2) {
                const fixedSet = (target.kind === 'union' || target.kind === 'stringLiteral' || target.kind === 'integerLiteral');
                const openSet = (target.kind === 'base' && (target.name === 'string' || target.name === 'integer' || target.name === 'uinteger'));
                if (openSet || fixedSet) {
                    // Check if we have a enum declaration.
                    const body = namespace.getChildren().find(node => node.kind === ts.SyntaxKind.ModuleBlock);
                    if (body !== undefined && ts.isModuleBlock(body)) {
                        const enumValues = this.getEnumValues(target);
                        const variableStatements = body.statements.filter((statement => ts.isVariableStatement(statement)));
                        if ((fixedSet && enumValues !== undefined && enumValues.length > 0 && variableStatements.length === enumValues.length) || (openSet && variableStatements.length > 0)) {
                            // Same length and all variable statement.
                            const enumValuesSet = enumValues ? new Set(enumValues) : undefined;
                            let isEnum = true;
                            const enumerations = [];
                            for (const variable of variableStatements) {
                                if (!ts.isVariableStatement(variable) || variable.declarationList.declarations.length !== 1) {
                                    isEnum = false;
                                    break;
                                }
                                const declaration = variable.declarationList.declarations[0];
                                if (!ts.isVariableDeclaration(declaration)) {
                                    isEnum = false;
                                    break;
                                }
                                const value = this.getEnumValue(declaration);
                                if (value === undefined) {
                                    isEnum = false;
                                    break;
                                }
                                if (enumValuesSet && !enumValuesSet.has(value)) {
                                    isEnum = false;
                                    break;
                                }
                                let propertyName = declaration.name.getText();
                                if (Visitor.PropertyRenames.has(name) && Visitor.PropertyRenames.get(name)?.has(propertyName)) {
                                    propertyName = Visitor.PropertyRenames.get(name).get(propertyName);
                                }
                                if (Visitor.PropertyFilters.has(name) && Visitor.PropertyFilters.get(name)?.has(propertyName)) {
                                    continue;
                                }
                                const entry = { name: propertyName, value: value };
                                this.fillDocProperties(variable, entry);
                                enumerations.push(entry);
                            }
                            if (isEnum) {
                                const type = enumValues
                                    ? this.getEnumBaseType(enumValues)
                                    : openSet ? target : undefined;
                                if (type !== undefined) {
                                    const enumeration = { name: name, type: type, values: enumerations };
                                    if (openSet && !fixedSet) {
                                        enumeration.supportsCustomValues = true;
                                    }
                                    // First fill the documentation from the namespace and then from the
                                    // type declaration.
                                    this.fillDocProperties(namespace, enumeration);
                                    if (enumeration.documentation === undefined) {
                                        this.fillDocProperties(namespace, enumeration);
                                    }
                                    return enumeration;
                                }
                            }
                        }
                    }
                }
            }
            // We have a single reference to another type. Treat is as an extend of
            // that structure
            if (target.kind === 'reference' && Visitor.Mixins.has(target.name)) {
                this.queueTypeInfo(target);
                const result = { name: name, mixins: [TypeInfo.asJsonType(target)], properties: [] };
                this.fillDocProperties(declaration, result);
                return result;
            }
            else {
                this.queueTypeInfo(target);
                // In the protocol document filter can't be a string but the implementation allows it for backwards compatibility
                if (name === 'DocumentSelector') {
                    return PreDefined.DocumentSelector;
                }
                const result = { name: name, type: TypeInfo.asJsonType(target) };
                this.fillDocProperties(declaration, result);
                return result;
            }
        }
        else if (typescripts_1.Symbols.isRegularEnum(symbol)) {
            const entries = [];
            const exports = this.typeChecker.getExportsOfModule(symbol);
            let enumBaseType = undefined;
            for (const item of exports) {
                const declaration = this.getDeclaration(item, ts.SyntaxKind.EnumMember);
                if (declaration === undefined || !ts.isEnumMember(declaration) || declaration.initializer === undefined) {
                    continue;
                }
                let value;
                if (ts.isNumericLiteral(declaration.initializer)) {
                    value = Number.parseInt(declaration.initializer.getText());
                    if (value >= 0 && enumBaseType === undefined) {
                        enumBaseType = 'uinteger';
                    }
                    else {
                        enumBaseType = 'integer';
                    }
                }
                else if (ts.isStringLiteral(declaration.initializer)) {
                    value = this.removeQuotes(declaration.initializer.getText());
                    enumBaseType = 'string';
                }
                if (value === undefined) {
                    continue;
                }
                const entry = { name: item.getName(), value: value };
                if (Visitor.PropertyFilters.has(name) && Visitor.PropertyFilters.get(name)?.has(entry.name)) {
                    continue;
                }
                this.fillDocProperties(declaration, entry);
                entries.push(entry);
            }
            const type = enumBaseType === undefined
                ? { kind: 'base', name: 'uinteger' }
                : { kind: 'base', name: enumBaseType };
            const result = { name: name, type: type, values: entries };
            if (name === 'SemanticTokenTypes' || name === 'SemanticTokenModifiers') {
                result.supportsCustomValues = true;
            }
            const declaration = this.getDeclaration(symbol, ts.SyntaxKind.EnumDeclaration);
            if (declaration !== undefined) {
                this.fillDocProperties(declaration, result);
            }
            return result;
        }
        else {
            const result = { name: name, properties: [] };
            this.fillProperties(result, symbol);
            const declaration = this.getFirstDeclaration(symbol);
            declaration !== undefined && this.fillDocProperties(declaration, result);
            return result;
        }
    }
    fillProperties(result, symbol) {
        // Using the type here to navigate the properties will result in folding
        // all properties since the type contains all inherited properties. So we go
        // over the symbol to make things work.
        if (symbol.members === undefined) {
            return;
        }
        symbol.members.forEach((member) => {
            if (!typescripts_1.Symbols.isProperty(member)) {
                return;
            }
            const declaration = this.getDeclaration(member, ts.SyntaxKind.PropertySignature);
            if (declaration === undefined || !ts.isPropertySignature(declaration) || declaration.type === undefined) {
                throw new Error(`Can't parse property ${member.getName()} of structure ${symbol.getName()}`);
            }
            const typeInfo = this.getTypeInfo(declaration.type);
            if (typeInfo === undefined) {
                throw new Error(`Can't parse property ${member.getName()} of structure ${symbol.getName()}`);
            }
            const isExperimentalProperty = (result.name === 'ServerCapabilities' && member.getName() === 'experimental' && typeInfo.kind === 'reference' && typeInfo.name === 'T');
            const property = isExperimentalProperty
                ? { name: member.getName(), type: TypeInfo.asJsonType({ kind: 'reference', name: 'LSPAny', symbol: typeInfo.symbol }) }
                : { name: member.getName(), type: TypeInfo.asJsonType(typeInfo) };
            if (typescripts_1.Symbols.isOptional(member)) {
                property.optional = true;
            }
            this.fillDocProperties(declaration, property);
            result.properties.push(property);
            if (!isExperimentalProperty) {
                this.queueTypeInfo(typeInfo);
            }
        });
    }
    splitIntersectionType(node) {
        let literal;
        const rest = [];
        const references = [];
        for (const element of node.types) {
            if (ts.isTypeLiteralNode(element)) {
                if (literal === undefined) {
                    literal = element;
                }
                else {
                    rest.push(element);
                }
            }
            else if (ts.isTypeReferenceNode(element)) {
                references.push(element);
            }
            else {
                rest.push(element);
            }
        }
        return { literal, references, rest };
    }
    getFirstDeclaration(symbol) {
        const declarations = symbol.getDeclarations();
        return declarations !== undefined && declarations.length > 0 ? declarations[0] : undefined;
    }
    getDeclaration(symbol, kind) {
        const declarations = symbol.getDeclarations();
        if (declarations === undefined) {
            return undefined;
        }
        for (const declaration of declarations) {
            if (declaration.kind === kind) {
                return declaration;
            }
        }
        return undefined;
    }
    getEnumValues(typeInfo) {
        if (typeInfo.kind === 'stringLiteral') {
            return [typeInfo.value];
        }
        if (typeInfo.kind === 'integerLiteral') {
            return [typeInfo.value];
        }
        if (typeInfo.kind !== 'union' || typeInfo.items.length === 0) {
            return undefined;
        }
        const first = typeInfo.items[0];
        const item = first.kind === 'stringLiteral' ? [first.kind, first.value] : (first.kind === 'integerLiteral' ? [first.kind, first.value] : undefined);
        if (item === undefined) {
            return undefined;
        }
        const kind = item[0];
        const result = [];
        result.push(item[1]);
        for (let i = 1; i < typeInfo.items.length; i++) {
            const info = typeInfo.items[i];
            if (info.kind !== kind) {
                return undefined;
            }
            if (info.kind !== 'integerLiteral' && info.kind !== 'stringLiteral') {
                return undefined;
            }
            result.push(info.value);
        }
        return result;
    }
    getEnumValue(declaration) {
        let enumValueNode;
        if (declaration.initializer !== undefined) {
            enumValueNode = declaration.initializer;
        }
        else if (declaration.type !== undefined && ts.isLiteralTypeNode(declaration.type)) {
            enumValueNode = declaration.type.literal;
        }
        if (enumValueNode === undefined) {
            return undefined;
        }
        if (ts.isNumericLiteral(enumValueNode) || (ts.isPrefixUnaryExpression(enumValueNode) && ts.isNumericLiteral(enumValueNode.operand))) {
            return Number.parseInt(enumValueNode.getText());
        }
        else if (ts.isStringLiteral(enumValueNode)) {
            return this.removeQuotes(enumValueNode.getText());
        }
        return undefined;
    }
    getEnumBaseType(values) {
        if (values.length === 0) {
            return undefined;
        }
        const first = values[0];
        if (typeof first === 'string') {
            return { kind: 'base', name: 'string' };
        }
        for (const value of values) {
            if (value < 0) {
                return { kind: 'base', name: 'integer' };
            }
        }
        return { kind: 'base', name: 'uinteger' };
    }
    removeQuotes(text) {
        const first = text[0];
        if ((first !== '\'' && first !== '"' && first !== '`') || first !== text[text.length - 1]) {
            return text;
        }
        return text.substring(1, text.length - 1);
    }
    fillDocProperties(node, value) {
        const filePath = node.getSourceFile().fileName;
        const fileName = path.basename(filePath);
        const tags = ts.getJSDocTags(node);
        const { since, deprecated } = this.getTags(tags);
        const proposed = (fileName.startsWith('proposed.') || tags.some((tag) => { return ts.isJSDocUnknownTag(tag) && tag.tagName.text === 'proposed'; })) ? true : undefined;
        value.documentation = this.getDocumentation(node);
        value.since = since;
        value.deprecated = deprecated;
        value.proposed = proposed;
    }
    getDocumentation(node) {
        const fullText = node.getFullText();
        const ranges = ts.getLeadingCommentRanges(fullText, 0);
        if (ranges !== undefined && ranges.length > 0) {
            const start = ranges[ranges.length - 1].pos;
            const end = ranges[ranges.length - 1].end;
            const text = fullText.substring(start, end).trim();
            if (text.startsWith('/**')) {
                const buffer = [];
                const lines = text.split(/\r?\n/);
                for (let i = 0; i < lines.length; i++) {
                    let noComment = lines[i].replace(/^\s*\/\*\*(.*)\s*\*\/\s*$|^(\s*\/\*\*)|^(\s*\*\/\s*)$|^(\s*\*)/, (_match, m1) => {
                        if (m1) {
                            return m1;
                        }
                        else {
                            return '';
                        }
                    });
                    // First line
                    if (i === 0 || i === lines.length - 1) {
                        noComment = noComment.trim();
                        if (noComment.length === 0) {
                            continue;
                        }
                    }
                    if (noComment.length > 0 && noComment[0] === ' ') {
                        noComment = noComment.substring(1);
                    }
                    buffer.push(noComment);
                }
                return buffer.join('\n');
            }
        }
        return undefined;
    }
    getTags(tags) {
        const result = {};
        for (const tag of tags) {
            if (tag.tagName.text === 'since' && typeof tag.comment === 'string') {
                result.since = tag.comment;
            }
            else if (tag.tagName.text === 'deprecated' && typeof tag.comment === 'string') {
                result.deprecated = tag.comment;
            }
            if (result.since !== undefined && result.deprecated !== undefined) {
                return result;
            }
        }
        return result;
    }
}
exports.default = Visitor;
_Visitor_currentSourceFile = new WeakMap();
Visitor.Mixins = new Set(['WorkDoneProgressParams', 'PartialResultParams', 'StaticRegistrationOptions', 'WorkDoneProgressOptions']);
Visitor.PropertyFilters = new Map([
    ['TraceValues', new Set(['Compact'])],
    ['ErrorCodes', new Set(['jsonrpcReservedErrorRangeStart', 'serverErrorStart', 'MessageWriteError', 'MessageReadError', 'PendingResponseRejected', 'ConnectionInactive', 'jsonrpcReservedErrorRangeEnd', 'serverErrorEnd'])],
    ['LSPErrorCodes', new Set(['lspReservedErrorRangeStart', 'lspReservedErrorRangeEnd'])]
]);
Visitor.PropertyRenames = new Map([
    ['MonikerKind', new Map([
            ['$export', 'export'],
            ['$import', 'import']
        ])]
]);
var PreDefined;
(function (PreDefined) {
    PreDefined.LSPAny = {
        name: 'LSPAny',
        type: {
            kind: 'or',
            items: [
                {
                    kind: 'reference',
                    name: 'LSPObject'
                },
                {
                    kind: 'reference',
                    name: 'LSPArray'
                },
                {
                    kind: 'base',
                    name: 'string'
                },
                {
                    kind: 'base',
                    name: 'integer'
                },
                {
                    kind: 'base',
                    name: 'uinteger'
                },
                {
                    kind: 'base',
                    name: 'decimal'
                },
                {
                    kind: 'base',
                    name: 'boolean'
                },
                {
                    kind: 'base',
                    name: 'null'
                }
            ]
        },
        documentation: 'The LSP any type.\nPlease note that strictly speaking a property with the value `undefined`\ncan\'t be converted into JSON preserving the property name. However for\nconvenience it is allowed and assumed that all these properties are\noptional as well.\n@since 3.17.0',
        since: '3.17.0'
    };
    PreDefined.LSPObject = {
        'name': 'LSPObject',
        'type': {
            'kind': 'map',
            'key': {
                'kind': 'base',
                'name': 'string'
            },
            'value': {
                'kind': 'reference',
                'name': 'LSPAny'
            }
        },
        'documentation': 'LSP object definition.\n@since 3.17.0',
        'since': '3.17.0'
    };
    PreDefined.LSPArray = {
        name: 'LSPArray',
        type: {
            kind: 'array',
            element: {
                kind: 'reference',
                name: 'LSPAny'
            }
        },
        documentation: 'LSP arrays.\n@since 3.17.0',
        since: '3.17.0'
    };
    PreDefined.DocumentSelector = {
        name: 'DocumentSelector',
        type: {
            kind: 'array',
            element: {
                kind: 'reference',
                name: 'DocumentFilter'
            }
        },
        documentation: 'A document selector is the combination of one or many document filters.\n\n@sample `let sel:DocumentSelector = [{ language: \'typescript\' }, { language: \'json\', pattern: \'**∕tsconfig.json\' }]`;\n\nThe use of a string as a document filter is deprecated @since 3.16.0.',
        since: '3.16.0.'
    };
})(PreDefined || (PreDefined = {}));
//# sourceMappingURL=visitor.js.map